local zoneOverhaulVersion = game.state.qt_zone_overhaul_version or -1
-- It's safe to assume that the player won't be doing any of these chats on the world map because you can't
-- use items on the world map.

-- Characters stranded in the east shouldn't be able to contact west npcs.
-- We don't consider characters to be cut off anywhere else, even if they're
-- trapped in a dark crypt or high peak or a farportal zone or demon plane or
-- whatever; those cases don't really have significant gameplay implications.
local westCond = function()
	local sq = game.player:hasQuest("start-sunwall")
	-- Player DEFINITELY has access to West, and if they're currently in East then they can get back to East.
	-- (You can reuse the Reknor farportal if you haven't given Tannen your orb yet.)
	if game.player:hasQuest("east-portal") then
		return true
	-- Player has not used Reknor portal yet, and isn't doing the East starting quest.
	elseif ((not sq) or sq:isCompleted("slazish")) and (not game.player:isQuestStatus("wild-wild-east", engine.Quest.DONE)) then
		return true
	else
	-- Player must be stuck in East in vanilla, but can reach West in zoneOverhaulVersion >= 12
		return ((not sq) or sq:isCompleted("slazish")) and zoneOverhaulVersion >= 12
	end
end

-- And characters stranded in the west shouldn't be able to contact east npcs.
local eastCond = function()
	-- Player has been through Reknor farportal.
	if game.player:isQuestStatus("wild-wild-east", engine.Quest.DONE) then
		-- Lazy shortcut: if they've used the Reknor farportal, the only way they won't have the Orb (and therefore
		-- definitely be able to access East) is if Tannen took it and they're therefore stranded in West.
		-- Actually this check isn't necessary at the moment because the only east npc we include is Zemekkys,
		-- whose quest has to be completed (removing him from the menu) to reach this line in the first place...
		return game.party:findInAllPartyInventoriesBy("define_as", "ORB_MANY_WAYS")
	else
		-- If they haven't been through the portal, they only have access to East if they started there and are still there.
		local q = game.player:hasQuest("start-sunwall")
		return q and not q:isCompleted("slazish")
	end
end

local alchemistCond = function()
	-- Zone Overhaul removes the alchemists
	if zoneOverhaulVersion >= 12 then
		return false
	end
	-- If the quest is done, there is no reason to speak to alchemists.
	return westCond() and not (game.player:isQuestStatus("brotherhood-of-alchemists", engine.Quest.DONE))
end

local elderCond = function()
	if westCond() then
		-- Ready to give, or talk about, staff of absorption.
		local q = game.player:hasQuest("staff-absorption")
		if q and (q:isCompleted("ambush-died") or q:isCompleted("survived-ukruk")) then
			return true
		else
			-- Ready to talk about Tannen.
			q = game.player:hasQuest("east-portal")
			return zoneOverhaulVersion < 12 and q and not q:isCompleted("talked-elder")
		end
	else
		return false
	end
end

local zigurCond = function()
	-- No reason to talk to her if we aren't antimagic. Or she's dead. Or we have recent Zone Overhaul.
	return westCond() and zoneOverhaulVersion < 12 and game.player:attr("forbid_arcane") and game.player:attr("forbid_arcane") >= 2 and not game.player:attr("qt-myssil-dead")
end

local zemekkysCond = function()
	if zoneOverhaulVersion >= 12 then return false end
	local q = game.player:hasQuest("west-portal")
	-- XXX BAD: we need to be in Gates of Morning to make the portal, so don't allow chatting with him through the orb in that case.
	-- This should be replaced with an onLevelLoad.
	-- (Also don't allow chatting with him after making the portal, since he's not in his house anymore)
	if q and q:isCompleted("gem") then
		return false
	else
		return eastCond()
	end
end

-- Notable characters you can't speak to directly:
-- Anything involving a shop. Just doesn't work because of the way shops work in this game. (Removing shops is on the QuickTome TODO, anyway)
-- Limmir and the apprentice mage - may be added in the future, at the moment there are some possible issues that make me not trust them.
-- Zemekkys only when ready to create the farportal back to West; relies on you being on his level. Not ideal...
-- Aeryn, because of the need to move her around. TODO: work around this with an onLevelLoad.
-- Tannen and Linaniil, because either one can teleport you (which we don't want people using as an escape) and Tannen can screw your Orb of Many Ways (which we can't allow to happen if you're in east) in the rest of his chats.
-- Melinda's father in Last Hope. Again, teleporting. With QuickTome Melinda Tweaks, there's no gameplay reason to talk to him, anyway.
-- Arena guy. More teleporting, and with the full QuickTome set he is removed anyway.
-- Anything in Sher'Tul Fortress. Most of it relies on you actually being on that level.
newChat{ id="welcome",
	text = [[Speak to whom?]],
	answers = {
		{"Agrimley the hermit", cond = alchemistCond, action=function() require("engine.Chat").new("alchemist-hermit", {name="Agrimley the Alchemist"}, game.player):invoke() end},
		{"Marus of Elvala", cond = alchemistCond, action=function() require("engine.Chat").new("alchemist-elvala", {name="Marus the Alchemist"}, game.player):invoke() end},
		{"Stire of Derth", cond = alchemistCond, action=function() require("engine.Chat").new("alchemist-derth", {name="Stire the Alchemist"}, game.player):invoke() end},
		{"Ungrol of Last Hope", cond = alchemistCond, action=function() require("engine.Chat").new("alchemist-last-hope", {name="Ungrol the Alchemist"}, game.player):invoke() end},
		{"Elder of Last Hope", cond = elderCond, action=function() require("engine.Chat").new("last-hope-elder", {name="Elder of Last Hope"}, game.player):invoke() end},
		{"Protector Myssil", cond = zigurCond, action=function() require("engine.Chat").new("myssil", {name="Protector Myssil"}, game.player):invoke() end},
		{"Zemekkys of the Sunwall", cond = zemekkysCond, action=function() require("engine.Chat").new("zemekkys", {name="Zemekkys"}, game.player):invoke() end},
		{"Ansaya the Psionic Tutor", action=function() require("engine.Chat").new("qt-trainer", {name="Ansaya the Psionic Tutor"}, game.player):invoke() end},
		{"[Cancel]"},
	}
}

return "welcome"
