local DamageType = require("engine.DamageType")
local Stats = require "engine.interface.ActorStats"
local function doFileChanges(self,data,fileChangesTable)
	local fChanges = fileChangesTable[data.file]
	if fChanges then
		for _, obj in pairs(data.res) do
			if type(obj) == "table" then
				if fChanges.all then fChanges.all(obj) end
				local oChanges = fChanges[obj.define_as] or fChanges[obj.name]
				if oChanges then
					if oChanges.mut then oChanges.mut(obj) end
					if oChanges.set then
						for key,val in pairs(oChanges.set) do
							obj[key] = val
						end
					end
					if oChanges.remove then
						for _,key in pairs(oChanges.remove) do
							obj[key] = nil
						end
					end
				end
			end
		end
	end
end
-- Format:
-- [filename] = {
--   all = function(object),
--
--   [entity define_as OR entity name] = {
--     mut = function(object),
--     set = {
--       [field name] = {value},
--     },
--     remove = {
--       "[field name]"
--     },
--   },
-- }
--
-- The define_as takes precedence over the name, so if you have entries
-- for both ["ROGUE_PLIGHT"] and ["Rogue Plight"] only the first one
-- will be used.
--
-- all is an arbitrary function to run on every object. we're assuming
-- nobody is going to define an entity named "all"
--
-- mut is an arbitrary function to run with the object as its only
-- parameter, nice if you want to change a field in the entity without
-- overwriting it
--
-- set simply sets the entity's field to the value.
--
-- remove will set all the fields specified to nil.
--
-- All of these fields are optional.
--
-- The all function is run first, then the mut function, then set, then
-- remove.
--
-- example:
-- ["/data/general/objects/world-artifacts.lua"] = {
--   ["Summertide Phial"] = {
--     mut = function(obj) obj.wielder.lite = nil end,
--     set = {
--       encumber = 0.5,
--       cost = 100,
--     }
--     remove = { "max_power", "power_regen", "use_power" }
--   },
-- },
--
-- Q: Why do you allow using both the define_as and the name? Shouldn't you just use the name?
-- A: Many objects, such as Summertide Phial, actually lack a define_as. And many objects, such as most base NPCs, lack a name.
--    Objects that lack both, however, pretty much don't exist. And overlap between the fields is extremely unlikely.
local noGeneration = {remove={"rarity"}}
local massokChange = {
	-- find massok's dragonskull helm and replace it with helm of kroltar
	mut = function(obj)
		for k,v in pairs(obj) do
			if type(v) == "table" and v.__resolver and v.__resolver == "equip" and v[1] then
				for _,equip in pairs(v[1]) do
					if equip.defined=="DRAGON_SKULL" then equip.defined="HELM_KROLTAR" return end
				end
			end
		end
	end,
}
local fileChangesTable = {
	["/data/general/objects/world-artifacts.lua"] = {
		["Eden's Guile"] = noGeneration,
		["Ureslak's Molted Scales"] = noGeneration,
		["Wanderer's Rest"] = {
			remove = {
				"max_power",
				"power_regen",
				"use_talent",
			},
		},
		["Lightning Catcher"] = {
			mut = function(obj) -- constantly give the effects of 6 stacks - really players can easily keep it at 8-10 in the trunk version, and of course over a million in stable...
				obj.wielder.inc_stats = {
					[Stats.STAT_STR] = 6,
					[Stats.STAT_DEX] = 6,
					[Stats.STAT_MAG] = 6,
					[Stats.STAT_WIL] = 6,
					[Stats.STAT_CUN] = 6,
					[Stats.STAT_CON] = 6,
				}
				obj.wielder.inc_damage = {[DamageType.LIGHTNING] = 30 }
			end,
			remove = {
				"special_desc",
				"callbackOnTakeDamage",
				"callbackOnCrit",
			},
		},
		["Cuirass of the Thronesmen"] = {
			mut = function(obj)
				obj.wielder.combat_def = 30 -- from 20
				obj.wielder.combat_armor = 42 -- from 32
				obj.wielder.combat_armor_hardiness = 20 -- from 10
				obj.wielder.stun_immune = 0.6 -- from 0.4
			end,
		},
		["Wrap of Stone"] = {
			remove = {
				"max_power",
				"power_regen",
				"use_talent",
			},
		},
		["HELM_KROLTAR"] = {
			mut = function(obj)
				obj.wielder.combat_physresist = 12
				obj.wielder.combat_mentalresist = 12
				obj.wielder.combat_spellresist = 12
				obj.on_set_complete = function(self, who)
					self:specialSetAdd("skullcracker_mult", 1)
					self:specialSetAdd({"wielder","combat_spellresist"}, 3)
					self:specialSetAdd({"wielder","combat_mentalresist"}, 3)
					self:specialSetAdd({"wielder","combat_physresist"}, 3)
					game.logPlayer(who, "#GOLD#As the helm of Kroltar approaches the scale armour, they begin to fume and emit fire.")
				end
			end,
		},
		["DRAGON_SKULL"] = noGeneration,
	},
	["/data/general/objects/world-artifacts-far-east.lua"] = {
		["Threads of Fate"] = {
			remove = {
				"max_power",
				"power_regen",
				"use_talent",
			},
		},
		["GOEDALATH_ROCK"] = {
			remove = {
				"auto_pickup",
				"carrier",
				"on_pickup",
			},
		},
	},
	["/data/general/objects/boss-artifacts-far-east.lua"] = {
		["CROWN_ELEMENTS"] = {
			mut = function(obj)
				obj.wielder.inc_damage = { -- from 8 each
					[DamageType.FIRE] = 15,
					[DamageType.COLD] = 15,
					[DamageType.ACID] = 15,
					[DamageType.LIGHTNING] = 15,
				}
				obj.wielder.melee_project = { -- from 10 each
					[DamageType.FIRE] = 25,
					[DamageType.COLD] = 25,
					[DamageType.ACID] = 25,
					[DamageType.LIGHTNING] = 25,
				}
			end,
		},
	},
	["/data/zones/last-hope-graveyard/objects.lua"] = {
		["CELIA_HEART"] = {
			remove = {
				"max_power",
				"power_regen",
				"use_power",
				"use_sound",
			},
		},
	},
	["/data-ashes-urhrok/general/objects/world-artifacts.lua"] = {
		["The Black Ring"] = {
			set = {
				-- XXX: seems to require duplicating the string here...
				talent_on_spell = { {chance=10, talent="T_INVOKE_DARKNESS", level=3} },
			},
		},
	},
	["/data/zones/daikara/npcs.lua"] = {
		["MASSOK"] = massokChange,
	},
	["/data-qt-zone-overhaul/dreadfell-npcs.lua"] = {
		["MASSOK_QT"] = massokChange,
	},
	["/data/general/npcs/horror.lua"] = {
		["Ak'Gishil"] = {
			set = {
				on_added_to_level = function(self)
					self.blades = 0
					self.summoned_distort = 1 -- dont try to summon blade of distorted time anymore
				end,
			},
			mut = function(obj)
				table.insert(obj,resolvers.drops{chance=100, nb=1, {defined="RIFT_SWORD"}})
			end,
		},
		["ANIMATED_BLADE"] = {
			-- Do not summon unique longswords.
			mut = function(obj)
				for k,v in pairs(obj) do
					if type(v) == "table" and v.__resolver and v.__resolver == "equip" and v[1] then
						v[1][1].special = function(e) return not e.unique end
					end
				end
			end,
		},
	},
}

class:bindHook("Entity:loadList", function(self, data)
	doFileChanges(self,data,fileChangesTable)
end)
