local DamageType = require("engine.DamageType")
local ActorTalents = require("engine.interface.ActorTalents")
local function doFileChanges(self,data,fileChangesTable,allFunc)
	local fChanges = fileChangesTable[data.file]
	if fChanges or allFunc then
		for _, obj in pairs(data.res) do
			if type(obj) == "table" then
				if fChanges then
					if fChanges.all then fChanges.all(obj) end
					local oChanges = fChanges[obj.define_as] or fChanges[obj.name]
					if oChanges then
						if oChanges.mut then oChanges.mut(obj) end
						if oChanges.set then
							for key,val in pairs(oChanges.set) do
								obj[key] = val
							end
						end
						if oChanges.remove then
							for _,key in pairs(oChanges.remove) do
								obj[key] = nil
							end
						end
					end
				end
				if allFunc then
					allFunc(obj)
				end
			end
		end
	end
end
-- Format:
-- [filename] = {
--   all = function(object),
--
--   [entity define_as OR entity name] = {
--     mut = function(object),
--     set = {
--       [field name] = {value},
--     },
--     remove = {
--       "[field name]"
--     },
--   },
-- }
--
-- The define_as takes precedence over the name, so if you have entries
-- for both ["ROGUE_PLIGHT"] and ["Rogue Plight"] only the first one
-- will be used.
--
-- all is an arbitrary function to run on every object. we're assuming
-- nobody is going to define an entity named "all"
--
-- mut is an arbitrary function to run with the object as its only
-- parameter, nice if you want to change a field in the entity without
-- overwriting it
--
-- set simply sets the entity's field to the value.
--
-- remove will set all the fields specified to nil.
--
-- All of these fields are optional.
--
-- The all function is run first, then the mut function, then set, then
-- remove.
--
-- example:
-- ["/data/general/objects/world-artifacts.lua"] = {
--   ["Summertide Phial"] = {
--     mut = function(obj) obj.wielder.lite = nil end,
--     set = {
--       encumber = 0.5,
--       cost = 100,
--     }
--     remove = { "max_power", "power_regen", "use_power" }
--   },
-- },
--
-- Q: Why do you allow using both the define_as and the name? Shouldn't you just use the name?
-- A: Many objects, such as Summertide Phial, actually lack a define_as. And many objects, such as most base NPCs, lack a name.
--    Objects that lack both, however, pretty much don't exist. And overlap between the fields is extremely unlikely.
local function removeSightStuff(obj)
	local itemFields = {wielder=true,carrier=true,imbue_powers=true}
	local itemProps = {"blind_fight","blind_immune","lite","infravision","see_invisible","see_stealth","sight","esp","esp_range"}
	for k,v in pairs(obj) do
		if type(v) == "table" and v.__resolver and v.__resolver == "talents" then
			if v[1] and type(v[1]) == "table" then
				if v[1][ActorTalents.T_INVISIBILITY] then
					v[1][ActorTalents.T_INVISIBILITY] = nil
				end
				if v[1][ActorTalents.T_DARKNESS] then
					v[1][ActorTalents.T_DARKNESS] = nil
				end
			end
		elseif itemFields[k] then
			for _,prop in ipairs(itemProps) do
				v[prop] = nil
			end
		end
	end
end
local cutArtifactProps = {
		["immune blind"] = {remove = {"rarity"}},
		["see invisible"] = {remove = {"rarity"}},
		["infravision radius"] = {remove = {"rarity"}},
		["lite radius"] = {remove = {"rarity"}},
		["orc telepathy"] = {remove = {"rarity"}},
		["dragon telepathy"] = {remove = {"rarity"}},
		["demon telepathy"] = {remove = {"rarity"}},
	}
local noGeneration = {remove={"rarity"}}
local removeBadTalents = {mut = function(obj)
	for k,v in pairs(obj) do
		if type(v) == "table" and v.__resolver and v.__resolver == "talents" then
			if v[1] and type(v[1]) == "table" then
				if v[1][ActorTalents.T_INVISIBILITY] then
					v[1][ActorTalents.T_INVISIBILITY] = nil
				end
				if v[1][ActorTalents.T_DARKNESS] then
					v[1][ActorTalents.T_DARKNESS] = nil
				end
			end
		end
	end
end}
local removeCharm = function(obj)
	for k,v in pairs(obj) do
		if type(v) == "table" and v.__resolver and v.__resolver == "charmt" then
			obj[k] = nil
		end
	end
end
local fileChangesTable = {
	-- OBJECTS WE MAY WANT TO REMOVE, BUT CURRENTLY DON'T:
	-- Spectral Blade - silence on crit maybe saves it?
	-- Eye of the Forest
	-- Taint of Telepathy
	-- *All* divinations/telepathy in general (planned for future versions)

	-- OBJECTS WITH PROPERTIES WE KIND OF WANT TO REMOVE, BUT CURRENTLY DON'T:
	-- Star, Moon, Eye of the Wyrm, that garbage mummy set, and Unsetting Sun have light radius/blindness immunity on set complete,
	-- but if we overwrite the onSetComplete functions then
	-- we'd be incompatible with any other addon that touches those. Possibly future ToME4 versions too.
	-- Lightbringer's Wand and Dawn's Blade are similar.

	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	-- EGOS
	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	["/data/general/objects/random-artifacts/generic.lua"] = cutArtifactProps,
	["/data/general/objects/random-artifacts/ranged.lua"] = cutArtifactProps,
	["/data/general/objects/random-artifacts.lua"] = cutArtifactProps,
	["/data/general/objects/egos/amulets.lua"] = {
		["starlit "] = noGeneration, -- some crap resistances are not enough
		[" of vision"] = noGeneration,
	},
	["/data/general/objects/egos/armor.lua"] = {
		-- removing delving is a straight up item nerf...
		-- even removing radiant probably is. god, armor egos are crap and dumb
		["radiant "] = noGeneration,
		[" of delving"] = noGeneration,
	},
	["/data/general/objects/egos/digger.lua"] = {
		["miner's "] = noGeneration, -- the others with lite/infravision are fine even without them compared to other diggers
	},
	["/data/general/objects/egos/helm.lua"] = {
		-- the stat boosts on these are garbage for some reason, might as well toss both when bounder exists
		["champion's "] = noGeneration,
		["catburglar's "] = noGeneration,
		-- keeping miner's, no other lesser ego gives armor and it's not like it's worse than a helm of useless resists or stamina on hit
	},
	["/data/general/objects/egos/lite.lua"] = {
		["bright "] = noGeneration,
		[" of clarity"] = noGeneration,
		["watchleader's "] = noGeneration, -- THEre is too much random crap with trACK ON it
	},
	-- lots of sets with blind immune etc., but if we mess with those, we could hurt other addons, again.
	-- ["/data/general/objects/egos/mindstars.lua"]
	["/data/general/objects/egos/rings.lua"] = {
		[" of sensing"] = noGeneration,
	},
	["/data/general/objects/egos/staves.lua"] = {
		[" of illumination"] = noGeneration, -- yeah, we changed the talent, but who is going to care about getting it as a charm, honestly.
	},
	["/data/general/objects/egos/wands-powers.lua"] = {
		[" of clairvoyance"] = noGeneration,
	},
	["/data/general/objects/egos/wizard-hat.lua"] = {
		[" of the sentry"] = noGeneration,
		[" of the Brotherhood"] = {
			mut = removeCharm,
		},
	},
	["/data/general/objects/egos/gloves.lua"] = {
		[" of the nighthunter"] = noGeneration, -- Inferior to archery without track.
	},

	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	-- ARTIFACTS
	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	["/data/general/objects/world-artifacts.lua"] = {
		["Burning Star"] = noGeneration,
		["Eye of the Forest"] = {
			remove = {
				"max_power",
				"power_regen",
				"use_talent",
			},
		},
		["Fortune's Eye"] = { -- This is still OKish without track.
			remove = {
				"max_power",
				"power_regen",
				"use_talent",
			},
		},
		["Helm of the Dwarven Emperors"] = noGeneration,
		["Merkul's Second Eye"] = noGeneration, -- I'm tempted to leave this item in just because of how amusingly goofy it is, but nah.
		["Mummified Egg-sac of Ungolë"] = noGeneration,
		["Quiver of the Sun"] = {
			-- Lite light is the same as light now, so don't confuse players by pretending it isn't.
			mut = function(obj) obj.combat.damtype = DamageType.LIGHT end,
		},
		["Summertide Phial"] = { -- Really borderline, but it does have healing mod and a big resistance number on it. Leaving it in for now.
			remove = {
				"max_power",
				"power_regen",
				"use_power",
			},
		},
		["Umbraphage"] = { -- We're removing a LOT of stuff here, but it's still relatively good for some characters so leaving it in for now.
			remove = {
				"act",
				"max_power",
				"on_wear",
				"on_takeoff",
				"power_regen",
				"special_desc",
				"use_power",
			},
		},
		["Unflinching Eye"] = noGeneration, -- Its gimmicks really don't work for us.
		["Yaldan Baoth"] = noGeneration, -- Stat boost is redundant with Crown of Command, sight gimmick does very little now.
	},
	["/data/general/objects/boss-artifacts-maj-eyal.lua"] = {
		-- these items might still be useful for their dumb actives, but...those actives are dumb anyway! remove them
		["Wintertide Phial"] = noGeneration,
		["Choker of Dread"] = noGeneration,
		["Withering Orbs"] = noGeneration,
	},

	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	-- OTHER ITEMS
	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	["/data/general/objects/gem.lua"] = {
		["ALCHEMIST_GEM_AMETRINE"] = {
			mut = function(obj)
				-- Lite damage no longer does anything, so change it to light damage.
				obj.alchemist_bomb.splash = {type="LIGHT", dam=10}
			end,
		},
		["ALCHEMIST_GEM_PEARL"] = {
			mut = function(obj)
				-- Lite damage no longer does anything, so give 20% chance to blind instead.
				obj.alchemist_bomb.splash = {type="ITEM_LIGHT_BLIND", dam=20}
			end,
		},
		["GEM_TURQUOISE"] = {
			mut = function(obj)
				-- better powers welcome
				local powers = {max_life = 80}
				obj.imbue_powers = powers
				obj.wielder = powers
			end,
		},
		["GEM_CITRINE"] = {
			mut = function(obj)
				-- better powers welcome
				local powers = {healing_factor = 0.1}
				obj.imbue_powers = powers
				obj.wielder = powers
			end,
		},
	},
	["/data/general/objects/scrolls.lua"] = {
		["sun infusion"] = noGeneration, -- this is garbage even in the base game lol
		["vision rune"] = noGeneration,
		--["invisibility rune"] = noGeneration, -- For the moment, we just stop monsters from using these. XXX: We don't stop monsters from generating with them, though...
	},

	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	-- NPC DROP CHANGES
	--------------------------------------------------------------------------
	--------------------------------------------------------------------------
	["/data/zones/heart-gloom/npcs.lua"] = {
		["WITHERING_THING"] = {
			mut = function(obj)
				for k,v in pairs(obj) do
					if type(v) == "table" and v.__resolver then
						if v.__resolver == "drops" then 
						if v[1] and v[1][1] and v[1][1].defined and v[1][1].defined == "WITHERING_ORBS" then
							v[1][1].defined = "EYE_OF_THE_DREAMING_ONE"
						end
						end
					end
				end
			end,
		}
	},
	["/data/zones/blighted-ruins/npcs.lua"] = {
		["HALF_BONE_GIANT"] = {
			mut = function(obj)
				for k,v in pairs(obj) do
					if type(v) == "table" and v.__resolver then
						if v.__resolver == "drops" then 
						if v[1] and v[1][1] and v[1][1].defined and v[1][1].defined == "WINTERTIDE_PHIAL" then
							-- possibly replace with Skullcleaver?
							v[1][1].random_art_replace={chance=100}
						end
						end
					end
				end
			end,
		}
	},
	["/data/zones/dreadfell/npcs.lua"] = {
		["THE_MASTER"] = {
			mut = function(obj)
				if obj.equipment and type(obj.equipment[1]) == "table" then
					for k,v in pairs(obj.equipment[1]) do
						if type(v) == "table" and v.defined == "AMULET_DREAD" then
							v.random_art_replace={chance=100}
						end
					end
				end
			end,
		}
	},
}

class:bindHook("ToME:load", function(self, data)

	-- These two make vision symmetric
	core.fov.set_algorithm("large_ass")
	core.fov.set_actor_vision_size(1.0)

	-- This one makes vision better
	core.fov.set_permissiveness(1.0)

	require("engine.interface.ActorTemporaryEffects"):loadDefinition("/data-qt-sight-overhaul/timed_effects.lua")
	ActorTalents:loadDefinition("/data-qt-sight-overhaul/talentChanges.lua")

	--
	-- BELOW THIS LINE: stuff that should probably be somewhere else
	--

	-- fix abyssal shroud
	if DamageType.dam_def.ABYSSAL_SHROUD then
		DamageType.dam_def.ABYSSAL_SHROUD.projector = function(src, x, y, type, dam, state)
			state = DamageType.initState(state)
			DamageType.useImplicitCrit(src, state)

			local target = game.level.map(x, y, engine.Map.ACTOR)
			local reapplied = false
			if target then
				-- silence the apply message it if the target already has the effect
				for eff_id, p in pairs(target.tmp) do
					local e = target.tempeffect_def[eff_id]
					if e.desc == "Abyssal Shroud" then
						reapplied = true
					end
				end
				target:setEffect(target.EFF_ABYSSAL_SHROUD, 2, {power=dam.power, lite=dam.lite, apply_power=src:combatSpellpower(), min_dur=1}, reapplied)
				DamageType:get(DamageType.DARKNESS).projector(src, x, y, DamageType.DARKNESS, dam.dam, state)
			end
		end
	end

	-- Remove divinations and survival from birth options. This is important to get rid of class mastery bonuses and mastery/perfection egos.
	local birth = require("engine.Birther")
	for _,descriptor in ipairs(birth.birth_descriptor_def) do
		if type(descriptor) == "table" and descriptor.type == "subclass" and descriptor.talents_types and type(descriptor.talents_types) == "table" then
			descriptor.talents_types["cunning/survival"] = nil
			descriptor.talents_types["spell/divination"] = nil
		end
	end
end)

class:bindHook("Entity:loadList", function(self, data)
	doFileChanges(self,data,fileChangesTable,removeSightStuff)
end)
