local _M = loadPrevious(...)

-- New item drop rules.
--
--------------------------------------------------------------------------------
-- Normal drops
--------------------------------------------------------------------------------
-- Money is always 7%.
-- Lore is always 2%.
-- Uniques are always 3%.
--
-- Basics: We want this to decrease as level increases as it's the "worst"
--         non-fixed (lore, money) result.
-- Double greater: We want this to increase as level increases as it's the
--         "best" non-fixed (lore, money) result.
-- Ego, double_ego, greater, normal: We want each of these to go up (as they
--         replace the next-"best" option) and then down (as the next-"best"
--         option replaces them).
--
-- To accomplish this, I use a relatively simple set of rules:
-- - The chance for the item to be basic is set to 20*(2/3)^(level-1)%, rounded
--   to the nearest integer for ease of use.
--
-- - The remaining chance out of 100 belongs to egos.
--
-- - Let the egos correspond to the 5 levels: level 1 corresponds to ego, level 2 to double_ego, etc.
--
-- - The chance for egos is distributed among all eligible egos such that the
--   matching ego for the level (e.g. double_ego for level 2) has the highest
--   chance, the adjacent egos have 1/2 the chance each, the egos adjacent to
--   those have 1/4 chance each, the next 1/8 chance, etc.
--
-- - At level 1, only ego and double_ego are eligible and the rest are 0. At
--   level 2, only ego, double_ego and greater are eligible. This is a holdover
--   from the original tome4 drops, and not very pretty, but oh well.
--
-- This approach has a few weaknesses mathematically (in particular, you can't
-- easily extend it below level 1 or above level 5) but is much better than the
-- original drop table since the original normal drop table has no consistent
-- pattern at all. The weights also don't add up to 100 in the original drop
-- table which makes it rather difficult to read!
--
--------------------------------------------------------------------------------
-- Boss drops
--------------------------------------------------------------------------------
-- Money, basic, and lore are always 0% as they are basically noise.
-- Bosses usually drop between 3 and 5 items, with a slight bias towards 3.
-- We would like to drop,
-- on average, around half an extra unique per boss in addition to the
-- single unique that almost all bosses have defined. (Zone Overhaul removes a
-- lot of bosses and loot; this is some compensation). So we give uniques a 15%
-- chance at all levels.
--
-- For ego items, the same rules are used as for normal drops, except the level
-- is "transposed" up by one.
local drop_tables = {
	normal = {
		[1] = {
			uniques = 3,

			-- egos: 68% centered on ego
			double_greater =	0,
			greater_normal =	0,
			greater =		0,
			double_ego =	68/3,   -- 33% of egos
			ego =			68*2/3, -- 66% of egos

			basic = 20,
			money = 7,
			lore = 2,
		},
		[2] = {
			uniques = 3,

			-- egos: 75% centered on double ego
			double_greater =	0,
			greater_normal =	0,
			greater =		75/4,	-- 25% of egos
			double_ego =	75/2,	-- 50% of egos
			ego =			75/4,	-- 25% of egos

			basic = 13,
			money = 7,
			lore = 2,
		},
		[3] = {
			uniques = 3,

			-- egos: 79% centered on greater ego
			double_greater =	79/10,  -- 10% of egos
			greater_normal =	79/5,   -- 20% of egos
			greater =		79*2/5, -- 40% of egos
			double_ego =	79/5,   -- 20% of egos
			ego =			79/10,  -- 10% of egos

			basic = 9,
			money = 7,
			lore = 2,
		},
		[4] = {
			uniques = 3,

			-- egos: 82% centered on greater+normal
			-- the final numbers look weird but remember that
			-- greater_normal is exactly twice the weight of
			-- its neighbours, greater is twice the weight of
			-- double_ego, double_ego is twice the weight of
			-- ego and all the numbers still add up to exactly
			-- 100 (disregarding floating point errors anyway).
			double_greater =	656/38,  -- ~21.1% of egos
			greater_normal =	656/19,  -- ~42.1% of egos
			greater =		656/38,  -- ~21.1% of egos
			double_ego =	656/76,  -- ~10.5% of egos
			ego =			656/152, -- ~5.3% of egos

			basic = 6,
			money = 7,
			lore = 2,
		},
		[5] = {
			uniques = 3,

			-- egos: 84% centered on double greater
			double_greater =	1344/31,  -- ~51.6% of egos
			greater_normal =	1344/62,  -- ~25.8% of egos
			greater =		1344/124, -- ~12.9% of egos
			double_ego =	1344/248, -- ~6.5% of egos
			ego =			1344/496, -- ~3.2% of egos

			basic = 4,
			money = 7,
			lore = 2,
		},
	},
	store = {
		[1] = {
			uniques = 0.5,
			double_greater = 10,
			greater_normal = 15,
			greater = 25,
			double_ego = 45,
			ego = 10,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[2] = {
			uniques = 0.5,
			double_greater = 20,
			greater_normal = 18,
			greater = 25,
			double_ego = 35,
			ego = 8,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[3] = {
			uniques = 0.5,
			double_greater = 30,
			greater_normal = 22,
			greater = 25,
			double_ego = 25,
			ego = 6,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[4] = {
			uniques = 0.5,
			double_greater = 40,
			greater_normal = 30,
			greater = 25,
			double_ego = 20,
			ego = 4,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[5] = {
			uniques = 0.5,
			double_greater = 50,
			greater_normal = 30,
			greater = 25,
			double_ego = 10,
			ego = 0,
			basic = 0,
			money = 0,
			lore = 0,
		},
	},
	boss = {
		[1] = {
			uniques = 15,
			double_greater = 0,
			greater_normal = 0,
			greater = 85/4,
			double_ego = 85/2,
			ego = 85/4,

			basic = 0,
			money = 0,
			lore = 0,
		},
		[2] = {
			uniques = 15,
			double_greater =	85/10,
			greater_normal =	85/5,
			greater =		85*2/5,
			double_ego =	85/5,
			ego =			85/10,

			basic = 0,
			money = 0,
			lore = 0,
		},
		[3] = {
			uniques = 15,
			double_greater =	680/38,
			greater_normal =	680/19,
			greater =		680/38,
			double_ego =	680/76,
			ego =			680/152,

			basic = 0,
			money = 0,
			lore = 0,
		},
		[4] = {
			uniques = 15,
			double_greater =	1360/31,
			greater_normal =	1360/62,
			greater =		1360/124,
			double_ego =	1360/248,
			ego =			1360/496,

			basic = 0,
			money = 0,
			lore = 0,
		},
		[5] = {
			uniques = 15,
			double_greater =	2040/31,  -- ~77.4% of egos
			greater_normal =	1360/124, -- ~12.9% of egos
			greater =		1360/248, -- ~6.5% of egos
			double_ego =	1360/496, -- ~3.2% of egos
			ego = 0,

			basic = 0,
			money = 0,
			lore = 0,
		},
	},
}
local orig_drop_tables = {
	normal = {
		[1] = {
			uniques = 0.5,
			double_greater = 0,
			greater_normal = 0,
			greater = 0,
			double_ego = 20,
			ego = 45,
			basic = 38,
			money = 7,
			lore = 2,
		},
		[2] = {
			uniques = 0.7,
			double_greater = 0,
			greater_normal = 0,
			greater = 10,
			double_ego = 35,
			ego = 30,
			basic = 41,
			money = 8,
			lore = 2.5,
		},
		[3] = {
			uniques = 1,
			double_greater = 10,
			greater_normal = 15,
			greater = 25,
			double_ego = 25,
			ego = 25,
			basic = 10,
			money = 8.5,
			lore = 2.5,
		},
		[4] = {
			uniques = 1.1,
			double_greater = 15,
			greater_normal = 35,
			greater = 25,
			double_ego = 20,
			ego = 5,
			basic = 5,
			money = 8,
			lore = 3,
		},
		[5] = {
			uniques = 1.2,
			double_greater = 35,
			greater_normal = 30,
			greater = 20,
			double_ego = 10,
			ego = 5,
			basic = 5,
			money = 8,
			lore = 3,
		},
	},
	store = {
		[1] = {
			uniques = 0.5,
			double_greater = 10,
			greater_normal = 15,
			greater = 25,
			double_ego = 45,
			ego = 10,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[2] = {
			uniques = 0.5,
			double_greater = 20,
			greater_normal = 18,
			greater = 25,
			double_ego = 35,
			ego = 8,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[3] = {
			uniques = 0.5,
			double_greater = 30,
			greater_normal = 22,
			greater = 25,
			double_ego = 25,
			ego = 6,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[4] = {
			uniques = 0.5,
			double_greater = 40,
			greater_normal = 30,
			greater = 25,
			double_ego = 20,
			ego = 4,
			basic = 0,
			money = 0,
			lore = 0,
		},
		[5] = {
			uniques = 0.5,
			double_greater = 50,
			greater_normal = 30,
			greater = 25,
			double_ego = 10,
			ego = 0,
			basic = 0,
			money = 0,
			lore = 0,
		},
	},
	boss = {
		[1] = {
			uniques = 3,
			double_greater = 0,
			greater_normal = 0,
			greater = 5,
			double_ego = 45,
			ego = 45,
			basic = 0,
			money = 4,
			lore = 0,
		},
		[2] = {
			uniques = 4,
			double_greater = 0,
			greater_normal = 8,
			greater = 15,
			double_ego = 40,
			ego = 35,
			basic = 0,
			money = 4,
			lore = 0,
		},
		[3] = {
			uniques = 5,
			double_greater = 10,
			greater_normal = 22,
			greater = 25,
			double_ego = 25,
			ego = 20,
			basic = 0,
			money = 4,
			lore = 0,
		},
		[4] = {
			uniques = 6,
			double_greater = 40,
			greater_normal = 30,
			greater = 25,
			double_ego = 20,
			ego = 0,
			basic = 0,
			money = 4,
			lore = 0,
		},
		[5] = {
			uniques = 7,
			double_greater = 50,
			greater_normal = 30,
			greater = 25,
			double_ego = 10,
			ego = 0,
			basic = 0,
			money = 4,
			lore = 0,
		},
	},
}

local loot_mod = {
	uvault = { -- Uber vault
		uniques = 25,
		double_greater = 4,
		greater_normal = 2.5,
		greater = 1.5,
		double_ego = 0,
		ego = 0,
		basic = 0,
		money = 0,
		lore = 0,
		material_mod = 1,
	},
	gvault = { -- Greater vault
		uniques = 6,
		double_greater = 1.5,
		greater_normal = 1.5,
		greater = 1.5,
		double_ego = 1,
		ego = 0,
		basic = 0,
		money = 0,
		lore = 0,
		material_mod = 1,
	},
	vault = { -- Default vault
		uniques = 3,
		double_greater = 1.5,
		greater_normal = 2,
		greater = 2,
		double_ego = 1.5,
		ego = 0,
		basic = 0,
		money = 0,
		lore = 0,
		material_mod = 1,
	},
}

local default_drops = function(zone, level, what, forceOriginal)
	if zone.default_drops then return zone.default_drops end
	local lev = util.bound(math.ceil(zone:level_adjust_level(level, "object") / 10), 1, 5)
	print("[TOME ENTITY FILTER] making default loot table for", what, lev)
	return (game:isCampaign("Maj'Eyal") and not forceOriginal) and table.clone(drop_tables[what][lev]) or table.clone(orig_drop_tables[what][lev])
end

-- Replace entirely
function _M:defaultEntityFilter(zone, level, type)
	if type ~= "object" then return end

	-- By default we dont apply special filters, but we always provide one so that entityFilter is called
	return {
		tome = default_drops(zone, level, "normal"),
	}
end
-- Have to replace this entirely too...stupid upvalues!
function _M:entityFilterAlter(zone, level, type, filter)
	if type ~= "object" then return filter end

	if filter.force_tome_drops or (not filter.tome and not filter.defined and not filter.special and not filter.unique and not filter.ego_chance and not filter.ego_filter and not filter.no_tome_drops) then
		filter = table.clone(filter)
		filter.tome = default_drops(zone, level, filter.tome_drops or "normal")
	end

	if filter.tome then
		local t = (filter.tome == true) and default_drops(zone, level, "normal") or filter.tome
		filter.tome = nil

		if filter.tome_mod then
			t = table.clone(t)
			if _G.type(filter.tome_mod) == "string" then filter.tome_mod = loot_mod[filter.tome_mod] end
			for k, v in pairs(filter.tome_mod) do
--				print(" ***** LOOT MOD", k, v)
				t[k] = (t[k] or 0) * v
			end
		end

		-- If we request a specific type/subtype, we don't want categories that could make that not happen
		if filter.type or filter.subtype or filter.name then t.money = 0 t.lore = 0	end

		local u = t.uniques or 0
		local dg = u + (t.double_greater or 0)
		local ge = dg + (t.greater_normal or 0)
		local g = ge + (t.greater or 0)
		local de = g + (t.double_ego or 0)
		local e = de + (t.ego or 0)
		local m = e + (t.money or 0)
		local l = m + (t.lore or 0)
		local total = l + (t.basic or 0)

		local r = rng.float(0, total)
		if r < u then
			print("[TOME ENTITY FILTER] selected Uniques", r, u)
			filter.unique = true
			filter.not_properties = filter.not_properties or {}
			filter.not_properties[#filter.not_properties+1] = "lore"

		elseif r < dg then
			print("[TOME ENTITY FILTER] selected Double Greater", r, dg)
			filter.not_properties = filter.not_properties or {}
			filter.not_properties[#filter.not_properties+1] = "unique"
			filter.ego_chance={tries = { {ego_chance=100, properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source}, {ego_chance=100, properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source} } }

		elseif r < ge then
			print("[TOME ENTITY FILTER] selected Greater + Ego", r, ge)
			filter.not_properties = filter.not_properties or {}
			filter.not_properties[#filter.not_properties+1] = "unique"
			filter.ego_chance={tries = { {ego_chance=100, properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source}, {ego_chance=100, not_properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source} }}

		elseif r < g then
			print("[TOME ENTITY FILTER] selected Greater", r, g)
			filter.not_properties = filter.not_properties or {}
			filter.not_properties[#filter.not_properties+1] = "unique"
			filter.ego_chance={tries = { {ego_chance=100, properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source} } }

		elseif r < de then
			print("[TOME ENTITY FILTER] selected Double Ego", r, de)
			filter.not_properties = filter.not_properties or {}
			filter.not_properties[#filter.not_properties+1] = "unique"
			filter.ego_chance={tries = { {ego_chance=100, not_properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source}, {ego_chance=100, not_properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source} }}

		elseif r < e then
			print("[TOME ENTITY FILTER] selected Ego", r, e)
			filter.not_properties = filter.not_properties or {}
			filter.not_properties[#filter.not_properties+1] = "unique"
			filter.ego_chance={tries = { {ego_chance=100, not_properties={"greater_ego"}, power_source=filter.power_source, forbid_power_source=filter.forbid_power_source} } }

		elseif r < m then
			print("[TOME ENTITY FILTER] selected Money", r, m)
			filter.special = function(e) return e.type == "money" or e.type == "gem" end

		elseif r < l then
			print("[TOME ENTITY FILTER] selected Lore", r, l)
			filter.special = function(e) return e.lore and true or false end

		else
			print("[TOME ENTITY FILTER] selected basic", r, total)
			filter.not_properties = filter.not_properties or {}
			filter.not_properties[#filter.not_properties+1] = "unique"
			filter.ego_chance = -1000
		end
	end

	if filter.random_object then
		print("[TOME ENTITY FILTER] random object requested, removing ego chances")
		filter.ego_chance = -1000
	end

	-- By default we dont apply special filters, but we always provide one so that entityFilter is called
	return filter
end

-- Replace these functions entirely, without using the originals. We don't
-- want the normal backup guardian system to apply *at all*, so we do this
-- instead of changing onDie hooks or whatnot.
function _M:activateBackupGuardian(guardian, on_level, zonelevel, rumor, action)
-- No point doing anything here.
end

_M.qt_tier1QuestGranted = false
local orig_zoneCheckBackupGuardian = _M.zoneCheckBackupGuardian
function _M:zoneCheckBackupGuardian()
	if not game:isCampaign("Maj'Eyal") then return orig_zoneCheckBackupGuardian(self) end

	-- DUMB: This is an incredibly stupid place to put this, but it works, sorta. Grant the tier 1 zone quest, completely and utterly unrelated to backup guardians...
	if not ((self.qt_tier1QuestGranted and game.player:hasQuest("qt-tier1-zones")) or game.zone.infinite_dungeon or world.arena) then
		self.qt_tier1QuestGranted = true
		game.player:grantQuest("qt-tier1-zones")
		-- Q: Shouldn't you just do this in the on-grant?
		-- A: Nope! The on_grant doesn't work when you pick "restart with the same character". Yes, that's pretty WTF in itself...
		game.player:hasQuest("qt-tier1-zones"):makeOnLevelLoads()
		game.player:grantQuest("qt-zone-overhaul-info")
	end

	if not self.qt_allow_backup_guardians then return end
	local data = self.qt_allow_backup_guardians[game.zone.short_name]
	if data then
		-- Only adjust level once
		if data.new_level then
			game.zone.base_level = data.new_level
			if game.difficulty == game.DIFFICULTY_NIGHTMARE then
				game.zone.base_level_range = table.clone(game.zone.level_range, true)
				game.zone.specific_base_level.object = -10 -game.zone.base_level
				game.zone.base_level = game.zone.base_level * 1.35
			elseif game.difficulty == game.DIFFICULTY_INSANE then
				game.zone.base_level_range = table.clone(game.zone.level_range, true)
				game.zone.specific_base_level.object = -10 -game.zone.base_level
				game.zone.base_level = game.zone.base_level * 1.35 + 1
			elseif game.difficulty == game.DIFFICULTY_MADNESS then
				game.zone.base_level_range = table.clone(game.zone.level_range, true)
				game.zone.specific_base_level.object = -10 -game.zone.base_level
				game.zone.base_level = game.zone.base_level * 2.5 + 1
			end
			data.new_level = nil
		end

		-- Spawn any new guardians. Note the field name; this block won't run on "original" backup guardians even if they somehow made it into allow_backup_guardians
		if data.guardians_on_level[game.level.level] then
			for _,guardian in ipairs(data.guardians_on_level[game.level.level]) do
				-- I've made three "important" changes here:
				-- 1. max tries increased from 100 to 1000. Consider a level that's 95% walls; placement
				-- would fail 0.6% of the time with 100 tries, which is too high. The tries are not expensive
				-- and with 1000 you get acceptable numbers up to about 99% walls.
				-- 2. if it fails to pick a spot it'll add the entity anyway. sticking lithfengel in a wall, or even crashing, is
				-- actually better than making the game unwinnable
				-- 3. doesn't print the added guardian, since we are randomizing them now (would be an information leak). Of course,
				-- the game prints soooooooooooo many other information leaks...
				local m = game.zone:makeEntityByName(game.level, "actor", guardian.guardian)
				if m then
					local x, y = rng.range(0, game.level.map.w - 1), rng.range(0, game.level.map.h - 1)
					local tries = 0
					while not m:canMove(x, y) and tries < 1000 do
						x, y = rng.range(0, game.level.map.w - 1), rng.range(0, game.level.map.h - 1)
						tries = tries + 1
					end
					game.zone:addEntity(game.level, m, "actor", x, y)
				else
					print("WARNING: Backup Guardian not found: ", guardian.guardian)
				end
			end
			data.guardians_on_level[game.level.level] = nil
		end
		if next(data.guardians_on_level) == nil then
			-- Table is empty so we placed the last guardian and are done.
			self.allow_backup_guardians[game.zone.short_name] = nil
		end
	end
end

-- nop for save compat
function _M:qt_activateBackupGuardian(guardian, zone, on_level, zonelevel, rumor, action)

end

local orig_goneEast = _M.goneEast
function _M:goneEast()
	orig_goneEast(self)

	-- Don't require talking to Aeryn.
	game.player:grantQuest("orc-pride")
	game.player:grantQuest("strange-new-world")
	game.player:setQuestStatus("strange-new-world", engine.Quest.COMPLETED, "helped-fillarel")
	game.player:learnTalent(game.player.T_RELENTLESS_PURSUIT, true, 1, {no_unlearn=true})
	game.logPlayer(game.player, "#VIOLET#You have learned the talent Relentless Pursuit.")

	-- Make Vor Armoury
	game:onLevelLoad("wilderness-1", function(zone, level)
		local g = game.zone:makeEntityByName(level, "terrain", "VOR_ARMOURY")
		local spot = level:pickSpot{type="zone-pop", subtype="vor-armoury"}
		game.zone:addEntity(level, g, "terrain", spot.x, spot.y)
		game.nicer_tiles:updateAround(game.level, spot.x, spot.y)
		game.state:locationRevealAround(spot.x, spot.y)
	end)

	game:onLevelLoad("town-gates-of-morning-1",function(zone,level)
		-- Make west portal
		local g1 = game.zone:makeEntityByName(game.level, "terrain", "WEST_PORTAL")
		local g2 = game.zone:makeEntityByName(game.level, "terrain", "CWEST_PORTAL")

		local spot = game.level:pickSpot{type="pop-quest", subtype="farportal"}
		game.zone:addEntity(game.level, g1, "terrain", spot.x, spot.y)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+1, spot.y)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+2, spot.y)
		game.zone:addEntity(game.level, g1, "terrain", spot.x, spot.y+1)
		game.zone:addEntity(game.level, g2, "terrain", spot.x+1, spot.y+1)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+2, spot.y+1)
		game.zone:addEntity(game.level, g1, "terrain", spot.x, spot.y+2)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+1, spot.y+2)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+2, spot.y+2)
	end)

	game:onLevelLoad("town-last-hope-1",function(zone,level)
		-- Make east portal
		local g1 = game.zone:makeEntityByName(game.level, "terrain", "FAR_EAST_PORTAL")
		local g2 = game.zone:makeEntityByName(game.level, "terrain", "CFAR_EAST_PORTAL")
		local spot = game.level:pickSpot{type="pop-quest", subtype="farportal"}
		game.zone:addEntity(game.level, g1, "terrain", spot.x, spot.y)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+1, spot.y)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+2, spot.y)
		game.zone:addEntity(game.level, g1, "terrain", spot.x, spot.y+1)
		game.zone:addEntity(game.level, g2, "terrain", spot.x+1, spot.y+1)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+2, spot.y+1)
		game.zone:addEntity(game.level, g1, "terrain", spot.x, spot.y+2)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+1, spot.y+2)
		game.zone:addEntity(game.level, g1, "terrain", spot.x+2, spot.y+2)
	end)
end

-- For use by other QuickTome addons
_M.qt_zone_overhaul_version = 12

return _M
